import torch
import math
from typing import Type, Dict, Any, Tuple, Callable, List

from . import merge as merge
from .utils import isinstance_str, init_generator


def compute_merge(
    x: torch.Tensor,
    ca_tome_info: Dict[str, Any],
    cache_info: Dict[str, Any],
    inference_step,
) -> Tuple[Callable, ...]:
    original_h, original_w = ca_tome_info["size"]
    original_tokens = original_h * original_w
    downsample = int(math.ceil(math.sqrt(original_tokens // x.shape[1])))

    args = ca_tome_info["args"]

    if downsample <= args["max_downsample"]:
        w = int(math.ceil(original_w / downsample))
        h = int(math.ceil(original_h / downsample))
        similarity_threshold = args["similarity_threshold"]

        # Re-init the generator if it hasn't already been initialized or device has changed.
        if args["generator"] is None:
            args["generator"] = init_generator(x.device)
        elif args["generator"].device != x.device:
            args["generator"] = init_generator(x.device, fallback=args["generator"])

        # If the batch size is odd, then it's not possible for prompted and unprompted images to be in the same
        # batch, which causes artifacts with use_rand, so force it to be off.
        use_rand = False if x.shape[0] % 2 == 1 else args["use_rand"]
        m, u = merge.bipartite_soft_matching_random2d(
            x,
            w,
            h,
            args["sx"],
            args["sy"],
            similarity_threshold,
            cache_info,
            inference_step,
            no_rand=not use_rand,
            generator=args["generator"],
        )
    else:
        m, u = (merge.do_nothing, merge.do_nothing)

    m_a, u_a = (m, u) if args["merge_attn"] else (merge.do_nothing, merge.do_nothing)
    m_c, u_c = (
        (m, u) if args["merge_crossattn"] else (merge.do_nothing, merge.do_nothing)
    )
    m_m, u_m = (m, u) if args["merge_mlp"] else (merge.do_nothing, merge.do_nothing)

    return m_a, m_c, m_m, u_a, u_c, u_m  # Okay this is probably not very good


def make_ca_tome_block(block_class: Type[torch.nn.Module]) -> Type[torch.nn.Module]:
    """
    Make a patched class on the fly so we don't have to import any specific modules.
    This patch applies CA-ToMe to the forward function of the block.
    """

    class AdaptiveToMeBlock(block_class):
        # Save for unpatching later
        _parent = block_class

        def _forward(
            self, x: torch.Tensor, context: torch.Tensor = None
        ) -> torch.Tensor:
            m_a, m_c, m_m, u_a, u_c, u_m = compute_merge(
                x, self._ca_tome_info, self._cache_info, self.inference_step
            )

            # This is where the meat of the computation happens
            x = (
                u_a(
                    self.attn1(
                        m_a(self.norm1(x)),
                        context=context if self.disable_self_attn else None,
                    )
                )
                + x
            )
            x = u_c(self.attn2(m_c(self.norm2(x)), context=context)) + x
            x = u_m(self.ff(m_m(self.norm3(x)))) + x

            return x

    return AdaptiveToMeBlock


def make_diffusers_ca_tome_block(
    block_class: Type[torch.nn.Module],
) -> Type[torch.nn.Module]:
    """
    Make a patched class for a diffusers model.
    This patch applies CA-ToMe to the forward function of the block.
    """

    class AdaptiveToMeBlock(block_class):
        # Save for unpatching later
        _parent = block_class

        def forward(
            self,
            hidden_states,
            attention_mask=None,
            encoder_hidden_states=None,
            encoder_attention_mask=None,
            timestep=None,
            cross_attention_kwargs=None,
            class_labels=None,
        ) -> torch.Tensor:
            # (1) CA-ToMe
            m_a, m_c, m_m, u_a, u_c, u_m = compute_merge(
                hidden_states, self._ca_tome_info, self._cache_info, self.inference_step
            )

            if self.use_ada_layer_norm:
                norm_hidden_states = self.norm1(hidden_states, timestep)
            elif self.use_ada_layer_norm_zero:
                norm_hidden_states, gate_msa, shift_mlp, scale_mlp, gate_mlp = (
                    self.norm1(
                        hidden_states,
                        timestep,
                        class_labels,
                        hidden_dtype=hidden_states.dtype,
                    )
                )
            else:
                norm_hidden_states = self.norm1(hidden_states)

            # (2) CA-ToMe m_a
            norm_hidden_states = m_a(norm_hidden_states)

            # 1. Self-Attention
            cross_attention_kwargs = (
                cross_attention_kwargs if cross_attention_kwargs is not None else {}
            )
            attn_output = self.attn1(
                norm_hidden_states,
                encoder_hidden_states=(
                    encoder_hidden_states if self.only_cross_attention else None
                ),
                attention_mask=attention_mask,
                **cross_attention_kwargs,
            )
            if self.use_ada_layer_norm_zero:
                attn_output = gate_msa.unsqueeze(1) * attn_output

            # (3) CA-ToMe u_a
            hidden_states = u_a(attn_output) + hidden_states

            if self.attn2 is not None:
                norm_hidden_states = (
                    self.norm2(hidden_states, timestep)
                    if self.use_ada_layer_norm
                    else self.norm2(hidden_states)
                )
                # (4) CA-ToMe m_c
                norm_hidden_states = m_c(norm_hidden_states)

                # 2. Cross-Attention
                attn_output = self.attn2(
                    norm_hidden_states,
                    encoder_hidden_states=encoder_hidden_states,
                    attention_mask=encoder_attention_mask,
                    **cross_attention_kwargs,
                )
                # (5) CA-ToMe u_c
                hidden_states = u_c(attn_output) + hidden_states

            # 3. Feed-forward
            norm_hidden_states = self.norm3(hidden_states)

            if self.use_ada_layer_norm_zero:
                norm_hidden_states = (
                    norm_hidden_states * (1 + scale_mlp[:, None]) + shift_mlp[:, None]
                )

            # (6) CA-ToMe m_m
            norm_hidden_states = m_m(norm_hidden_states)

            ff_output = self.ff(norm_hidden_states)

            if self.use_ada_layer_norm_zero:
                ff_output = gate_mlp.unsqueeze(1) * ff_output

            # (7) CA-ToMe u_m
            hidden_states = u_m(ff_output) + hidden_states

            return hidden_states

    return AdaptiveToMeBlock


def hook_ca_tome_model(model: torch.nn.Module):
    """Adds a forward pre hook to get the image size. This hook can be removed with remove_patch."""

    def hook(module, args):
        module._ca_tome_info["size"] = (args[0].shape[2], args[0].shape[3])
        return None

    model._ca_tome_info["hooks"].append(model.register_forward_pre_hook(hook))


def apply_patch(
    model: torch.nn.Module,
    compute_on: List[int],
    r: float = 0.7,
    max_downsample: int = 1,
    sx: int = 2,
    sy: int = 2,
    use_rand: bool = True,
    merge_attn: bool = True,
    merge_crossattn: bool = False,
    merge_mlp: bool = False,
):
    """
    Patches a Stable Diffusion model with Cached Adaptive Token Merging (CA-ToMe) optimization.
    Apply this to the highest level Stable Diffusion object (i.e., it should have a .model.diffusion_model).

    Important Args:
     - model: A top level Stable Diffusion module to patch in place. Should have a ".model.diffusion_model"
     - compute_on: List[int] specifying the timesteps where token merging patterns should be recomputed.
                  These checkpoints control when CA-ToMe updates its merging strategy during inference.
     - r: float between -1 and 1, cosine similarity threshold that determines whether tokens can be merged.
          Higher values (e.g. 0.7) require tokens to be more similar for merging.

    Optional Args:
     - max_downsample [1, 2, 4, or 8]: Controls which layers get CA-ToMe optimization based on downsampling.
                                       1 = only base resolution layers (4/15 layers)
                                       8 = all layers (15/15 layers)
                                       Recommended values: 1 or 2 for best quality/speed trade-off.
     - sx, sy: Stride values for partitioning tokens into source/destination sets.
               Higher strides allow more aggressive merging but may reduce quality.
               Default (2, 2) works well for most cases. Does not need to divide image size.
     - use_rand: Whether to add randomness when selecting destination tokens.
                 Recommended True for better results, set False if experiencing artifacts.
     - merge_attn: Whether to merge tokens in self-attention layers (recommended True).
     - merge_crossattn: Whether to merge tokens in cross-attention layers (recommended False).
     - merge_mlp: Whether to merge tokens in MLP layers (recommended False).

    Note: We carefully reasoned about these optional parameters and chose these specific values to
    provide an optimal balance between performance and quality. Please refer to our paper for more
    details about our reasoning.
    """

    # Make sure the module is not currently patched
    # Remove any existing CA-ToMe patches before applying new ones
    remove_patch(model)

    # Check if using diffusers library or base model
    is_diffusers = isinstance_str(model, "DiffusionPipeline") or isinstance_str(
        model, "ModelMixin"
    )

    # Get reference to the core diffusion model
    if not is_diffusers:
        # Handle base Stable Diffusion model case
        if not hasattr(model, "model") or not hasattr(model.model, "diffusion_model"):
            raise RuntimeError(
                "Provided model was not a Stable Diffusion / Latent Diffusion model, as expected."
            )
        diffusion_model = model.model.diffusion_model
    else:
        # Handle diffusers library case
        diffusion_model = model.unet if hasattr(model, "unet") else model

    # Initialize CA-ToMe configuration info
    diffusion_model._ca_tome_info = {
        "size": None,  # Will be set during forward pass
        "hooks": [],  # Store hooks for cleanup
        "args": {
            "similarity_threshold": r,
            "max_downsample": max_downsample,
            "sx": sx,
            "sy": sy,
            "use_rand": use_rand,
            "generator": None,
            "merge_attn": merge_attn,
            "merge_crossattn": merge_crossattn,
            "merge_mlp": merge_mlp,
        },
    }

    # Add hook to capture input dimensions
    hook_ca_tome_model(diffusion_model)

    # Patch each transformer block with CA-ToMe functionality
    for _, module in diffusion_model.named_modules():
        if isinstance_str(module, "BasicTransformerBlock"):
            # Select appropriate CA-ToMe block implementation
            make_ca_tome_block_fn = (
                make_diffusers_ca_tome_block if is_diffusers else make_ca_tome_block
            )
            # Replace block class with CA-ToMe-enabled version
            module.__class__ = make_ca_tome_block_fn(module.__class__)
            module._ca_tome_info = diffusion_model._ca_tome_info

            # Initialize caching info for adaptive token merging
            module._cache_info = {
                "compute_on": compute_on,  # Timesteps to recompute merging
                "merge": None,  # Cached merge information
                "unmerge": None,  # Cached unmerge information
            }

            # Handle model-specific compatibility flags
            if not hasattr(module, "disable_self_attn") and not is_diffusers:
                module.disable_self_attn = False  # For SD 2.0 compatibility

            if not hasattr(module, "use_ada_layer_norm_zero") and is_diffusers:
                # For older diffusers versions
                module.use_ada_layer_norm = False
                module.use_ada_layer_norm_zero = False

    return model


def remove_patch(model: torch.nn.Module):
    """Removes a patch from a CA-ToMe Diffusion module if it was already patched."""
    # For diffusers
    model = model.unet if hasattr(model, "unet") else model

    for _, module in model.named_modules():
        if hasattr(module, "_ca_tome_info"):
            for hook in module._ca_tome_info["hooks"]:
                hook.remove()
            module._ca_tome_info["hooks"].clear()

        if module.__class__.__name__ == "AdaptiveToMeBlock":
            module.__class__ = module._parent

    return model



import torch
from typing import Tuple, Callable, Dict, Any
import time


def do_nothing(x: torch.Tensor):
    return x


def mps_gather_workaround(input, dim, index):
    if input.shape[-1] == 1:
        return torch.gather(
            input.unsqueeze(-1), dim - 1 if dim < 0 else dim, index.unsqueeze(-1)
        ).squeeze(-1)
    else:
        return torch.gather(input, dim, index)


def bipartite_soft_matching_random2d(
    metric: torch.Tensor,
    w: int,
    h: int,
    sx: int,
    sy: int,
    similarity_threshold: float,
    cache_info: Dict[str, Any],
    inference_step: int,
    no_rand: bool = False,
    generator: torch.Generator = None,
) -> Tuple[Callable, Callable]:
    """
    Performs bipartite soft matching to merge tokens in a 2D grid layout, with caching for efficiency.

    This function partitions tokens into source (src) and destination (dst) sets, then merges similar tokens
    from src into dst based on a similarity threshold. The dst tokens are selected by choosing one token
    randomly (or deterministically) within each (sx, sy) sized region of the grid.

    Args:
        metric (torch.Tensor): Tensor of shape [B, N, C] containing token features/embeddings used to compute similarities.
            B is batch size, N is number of tokens, C is feature dimension.
        w (int): Width of the token grid in number of tokens
        h (int): Height of the token grid in number of tokens
        sx (int): Stride in x-dimension for selecting dst tokens. Controls granularity of merging.
        sy (int): Stride in y-dimension for selecting dst tokens. Controls granularity of merging.
        similarity_threshold (float): Cosine similarity threshold in [-1, 1] that determines whether tokens can be merged.
            Higher values (e.g. 0.7) require tokens to be more similar for merging.
        cache_info (Dict[str, Any]): Dictionary containing caching configuration and state:
            - compute_on: List of timesteps where merging should be recomputed
            - merge: Cached merge information (None in first iteration)
            - unmerge: Cached unmerge information (None in first iteration)
        inference_step (int): Current inference timestep, used for cache management
        no_rand (bool, optional): If True, disable randomness and always select top-left token as dst.
            Defaults to False.
        generator (torch.Generator, optional): Random number generator for reproducibility.
            If None, will initialize new generator.

    Returns:
        Tuple[Callable, Callable]: Tuple of (merge_func, unmerge_func) that handle the token merging/unmerging.
            merge_func: Merges similar tokens according to computed matching
            unmerge_func: Reverses the merging operation to restore original token arrangement
            If similarity_threshold >= 1, returns (do_nothing, do_nothing).
    """

    # Return cached functions if not at a checkpoints
    if (
        inference_step not in cache_info["compute_on"]
        and cache_info["merge"] is not None
    ):
        return cache_info["merge"], cache_info["unmerge"]

    B, N, _ = metric.shape

    # Skip merging if threshold too high
    if similarity_threshold >= 1:
        return do_nothing, do_nothing

    # Handle MPS device edge case
    gather = mps_gather_workaround if metric.device.type == "mps" else torch.gather

    with torch.no_grad():
        # Calculate grid dimensions
        hsy, wsx = h // sy, w // sx

        # Initialize destination token selection
        if no_rand:
            # Deterministic: always select top-left token
            rand_idx = torch.zeros(hsy, wsx, 1, device=metric.device, dtype=torch.int64)
        else:
            # Random: select one token per grid cell randomly
            rand_idx = torch.randint(
                sy * sx,
                size=(hsy, wsx, 1),
                device=generator.device,
                generator=generator,
            ).to(metric.device)

        # Create token assignment buffer
        idx_buffer_view = torch.zeros(
            hsy, wsx, sy * sx, device=metric.device, dtype=torch.int64
        )
        # Mark destination tokens with -1
        idx_buffer_view.scatter_(
            dim=2,
            index=rand_idx,
            src=-torch.ones_like(rand_idx, dtype=rand_idx.dtype),
        )
        # Reshape to match image grid
        idx_buffer_view = (
            idx_buffer_view.view(hsy, wsx, sy, sx)
            .transpose(1, 2)
            .reshape(hsy * sy, wsx * sx)
        )

        # Handle non-divisible image dimensions
        if (hsy * sy) < h or (wsx * sx) < w:
            idx_buffer = torch.zeros(h, w, device=metric.device, dtype=torch.int64)
            idx_buffer[: (hsy * sy), : (wsx * sx)] = idx_buffer_view
        else:
            idx_buffer = idx_buffer_view

        # Sort to get dst|src ordering
        rand_idx = idx_buffer.reshape(1, -1, 1).argsort(dim=1)

        # Free temporary buffers
        del idx_buffer, idx_buffer_view

        # Split into source and destination indices
        num_dst = hsy * wsx
        a_idx = rand_idx[:, num_dst:, :]  # src tokens
        b_idx = rand_idx[:, :num_dst, :]  # dst tokens

        def split(x):
            """Split tensor into source and destination tokens"""
            C = x.shape[-1]
            src = gather(x, dim=1, index=a_idx.expand(B, N - num_dst, C))
            dst = gather(x, dim=1, index=b_idx.expand(B, num_dst, C))
            return src, dst

        # Compute cosine similarities
        metric = metric / metric.norm(dim=-1, keepdim=True)
        a, b = split(metric)
        scores = a @ b.transpose(-1, -2)

        # Find best matches greedily
        node_max, node_idx = scores.max(dim=-1)
        edge_idx = node_max.argsort(dim=-1, descending=True)[..., None]

        # Determine number of tokens to merge based on similarity threshold
        similarity_mask = node_max >= similarity_threshold
        similarity_count = similarity_mask.sum(dim=-1)
        r = similarity_count.max().item()

        # Split tokens into merged and unmerged sets
        unm_idx = edge_idx[..., r:, :]  # Unmerged tokens
        src_idx = edge_idx[..., :r, :]  # Tokens to be merged
        dst_idx = gather(node_idx[..., None], dim=-2, index=src_idx)

        # Store merged pairs for analysis
        pairs = torch.cat((src_idx, dst_idx), dim=-1)

        def merge(x: torch.Tensor, mode="mean") -> torch.Tensor:
            """Merge tokens according to computed CA-ToMe matching"""
            src, dst = split(x)
            n, t1, c = src.shape

            # Handle unmerged tokens
            unm = gather(src, dim=-2, index=unm_idx.expand(n, t1 - r, c))
            # Merge similar tokens
            src = gather(src, dim=-2, index=src_idx.expand(n, r, c))
            dst = dst.scatter_reduce(-2, dst_idx.expand(n, r, c), src, reduce=mode)

            return torch.cat([unm, dst], dim=1)

        def unmerge(x: torch.Tensor) -> torch.Tensor:
            """Reverse the merging operation"""
            unm_len = unm_idx.shape[1]
            unm, dst = x[..., :unm_len, :], x[..., unm_len:, :]
            _, _, c = unm.shape

            # Retrieve merged source tokens
            src = gather(dst, dim=-2, index=dst_idx.expand(B, r, c))

            # Reconstruct original token arrangement
            out = torch.zeros(B, N, c, device=x.device, dtype=x.dtype)
            # Place destination tokens
            out.scatter_(dim=-2, index=b_idx.expand(B, num_dst, c), src=dst)
            # Place unmerged source tokens
            out.scatter_(
                dim=-2,
                index=gather(
                    a_idx.expand(B, a_idx.shape[1], 1), dim=1, index=unm_idx
                ).expand(B, unm_len, c),
                src=unm,
            )
            # Place merged source tokens
            out.scatter_(
                dim=-2,
                index=gather(
                    a_idx.expand(B, a_idx.shape[1], 1), dim=1, index=src_idx
                ).expand(B, r, c),
                src=src,
            )

            return out

        # Cache the merge/unmerge functions
        cache_info["merge"] = merge
        cache_info["unmerge"] = unmerge

        return merge, unmerge



from typing import Any, Callable, Dict, List, Optional, Union, Tuple
from dataclasses import dataclass
from .patch import apply_patch

import torch

from diffusers.callbacks import MultiPipelineCallbacks, PipelineCallback
from diffusers.image_processor import PipelineImageInput
from diffusers.utils import (
    USE_PEFT_BACKEND,
    deprecate,
    BaseOutput,
    scale_lora_layers,
    unscale_lora_layers,
)
from diffusers.pipelines.stable_diffusion.pipeline_output import (
    StableDiffusionPipelineOutput,
)


@dataclass
class UNet2DConditionOutput(BaseOutput):
    """
    The output of [`UNet2DConditionModel`].

    Args:
        sample (`torch.Tensor` of shape `(batch_size, num_channels, height, width)`):
            The hidden states output conditioned on `encoder_hidden_states` input. Output of last layer of model.
    """

    sample: torch.Tensor = None


def register_inference_step(block, inference_step):
    block_names = [
        "CrossAttnDownBlock2D",
        "UNetMidBlock2DCrossAttn",
        "CrossAttnUpBlock2D",
    ]
    if block.__class__.__name__ in block_names:
        for attn in block.attentions:
            for transformer in attn.transformer_blocks:
                transformer.inference_step = inference_step


def rescale_noise_cfg(noise_cfg, noise_pred_text, guidance_rescale=0.0):
    """
    Rescale `noise_cfg` according to `guidance_rescale`. Based on findings of [Common Diffusion Noise Schedules and
    Sample Steps are Flawed](https://arxiv.org/pdf/2305.08891.pdf). See Section 3.4
    """
    std_text = noise_pred_text.std(
        dim=list(range(1, noise_pred_text.ndim)), keepdim=True
    )
    std_cfg = noise_cfg.std(dim=list(range(1, noise_cfg.ndim)), keepdim=True)
    # rescale the results from guidance (fixes overexposure)
    noise_pred_rescaled = noise_cfg * (std_text / std_cfg)
    # mix with the original results from guidance by factor guidance_rescale to avoid "plain looking" images
    noise_cfg = (
        guidance_rescale * noise_pred_rescaled + (1 - guidance_rescale) * noise_cfg
    )
    return noise_cfg


def retrieve_timesteps(
    scheduler,
    num_inference_steps: Optional[int] = None,
    device: Optional[Union[str, torch.device]] = None,
    timesteps: Optional[List[int]] = None,
    sigmas: Optional[List[float]] = None,
    **kwargs,
):
    """
    Calls the scheduler's `set_timesteps` method and retrieves timesteps from the scheduler after the call. Handles
    custom timesteps. Any kwargs will be supplied to `scheduler.set_timesteps`.

    Args:
        scheduler (`SchedulerMixin`):
            The scheduler to get timesteps from.
        num_inference_steps (`int`):
            The number of diffusion steps used when generating samples with a pre-trained model. If used, `timesteps`
            must be `None`.
        device (`str` or `torch.device`, *optional*):
            The device to which the timesteps should be moved to. If `None`, the timesteps are not moved.
        timesteps (`List[int]`, *optional*):
            Custom timesteps used to override the timestep spacing strategy of the scheduler. If `timesteps` is passed,
            `num_inference_steps` and `sigmas` must be `None`.
        sigmas (`List[float]`, *optional*):
            Custom sigmas used to override the timestep spacing strategy of the scheduler. If `sigmas` is passed,
            `num_inference_steps` and `timesteps` must be `None`.

    Returns:
        `Tuple[torch.Tensor, int]`: A tuple where the first element is the timestep schedule from the scheduler and the
        second element is the number of inference steps.
    """
    if timesteps is not None and sigmas is not None:
        raise ValueError(
            "Only one of `timesteps` or `sigmas` can be passed. Please choose one to set custom values"
        )
    if timesteps is not None:
        accepts_timesteps = "timesteps" in set(
            inspect.signature(scheduler.set_timesteps).parameters.keys()
        )
        if not accepts_timesteps:
            raise ValueError(
                f"The current scheduler class {scheduler.__class__}'s `set_timesteps` does not support custom"
                f" timestep schedules. Please check whether you are using the correct scheduler."
            )
        scheduler.set_timesteps(timesteps=timesteps, device=device, **kwargs)
        timesteps = scheduler.timesteps
        num_inference_steps = len(timesteps)
    elif sigmas is not None:
        accept_sigmas = "sigmas" in set(
            inspect.signature(scheduler.set_timesteps).parameters.keys()
        )
        if not accept_sigmas:
            raise ValueError(
                f"The current scheduler class {scheduler.__class__}'s `set_timesteps` does not support custom"
                f" sigmas schedules. Please check whether you are using the correct scheduler."
            )
        scheduler.set_timesteps(sigmas=sigmas, device=device, **kwargs)
        timesteps = scheduler.timesteps
        num_inference_steps = len(timesteps)
    else:
        scheduler.set_timesteps(num_inference_steps, device=device, **kwargs)
        timesteps = scheduler.timesteps
    return timesteps, num_inference_steps


def modify_unet(unet):
    def reg_new_forward(self):
        def forward(
            sample: torch.Tensor,
            timestep: Union[torch.Tensor, float, int],
            inference_step: int,
            encoder_hidden_states: torch.Tensor,
            class_labels: Optional[torch.Tensor] = None,
            timestep_cond: Optional[torch.Tensor] = None,
            attention_mask: Optional[torch.Tensor] = None,
            cross_attention_kwargs: Optional[Dict[str, Any]] = None,
            added_cond_kwargs: Optional[Dict[str, torch.Tensor]] = None,
            down_block_additional_residuals: Optional[Tuple[torch.Tensor]] = None,
            mid_block_additional_residual: Optional[torch.Tensor] = None,
            down_intrablock_additional_residuals: Optional[Tuple[torch.Tensor]] = None,
            encoder_attention_mask: Optional[torch.Tensor] = None,
            return_dict: bool = True,
        ) -> Union[UNet2DConditionOutput, Tuple]:
            r"""
            The [`UNet2DConditionModel`] forward method.

            Args:
                sample (`torch.Tensor`):
                    The noisy input tensor with the following shape `(batch, channel, height, width)`.
                timestep (`torch.Tensor` or `float` or `int`): The number of timesteps to denoise an input.
                encoder_hidden_states (`torch.Tensor`):
                    The encoder hidden states with shape `(batch, sequence_length, feature_dim)`.
                class_labels (`torch.Tensor`, *optional*, defaults to `None`):
                    Optional class labels for conditioning. Their embeddings will be summed with the timestep embeddings.
                timestep_cond: (`torch.Tensor`, *optional*, defaults to `None`):
                    Conditional embeddings for timestep. If provided, the embeddings will be summed with the samples passed
                    through the `self.time_embedding` layer to obtain the timestep embeddings.
                attention_mask (`torch.Tensor`, *optional*, defaults to `None`):
                    An attention mask of shape `(batch, key_tokens)` is applied to `encoder_hidden_states`. If `1` the mask
                    is kept, otherwise if `0` it is discarded. Mask will be converted into a bias, which adds large
                    negative values to the attention scores corresponding to "discard" tokens.
                cross_attention_kwargs (`dict`, *optional*):
                    A kwargs dictionary that if specified is passed along to the `AttentionProcessor` as defined under
                    `self.processor` in
                    [diffusers.models.attention_processor](https://github.com/huggingface/diffusers/blob/main/src/diffusers/models/attention_processor.py).
                added_cond_kwargs: (`dict`, *optional*):
                    A kwargs dictionary containing additional embeddings that if specified are added to the embeddings that
                    are passed along to the UNet blocks.
                down_block_additional_residuals: (`tuple` of `torch.Tensor`, *optional*):
                    A tuple of tensors that if specified are added to the residuals of down unet blocks.
                mid_block_additional_residual: (`torch.Tensor`, *optional*):
                    A tensor that if specified is added to the residual of the middle unet block.
                down_intrablock_additional_residuals (`tuple` of `torch.Tensor`, *optional*):
                    additional residuals to be added within UNet down blocks, for example from T2I-Adapter side model(s)
                encoder_attention_mask (`torch.Tensor`):
                    A cross-attention mask of shape `(batch, sequence_length)` is applied to `encoder_hidden_states`. If
                    `True` the mask is kept, otherwise if `False` it is discarded. Mask will be converted into a bias,
                    which adds large negative values to the attention scores corresponding to "discard" tokens.
                return_dict (`bool`, *optional*, defaults to `True`):
                    Whether or not to return a [`~models.unets.unet_2d_condition.UNet2DConditionOutput`] instead of a plain
                    tuple.

            Returns:
                [`~models.unets.unet_2d_condition.UNet2DConditionOutput`] or `tuple`:
                    If `return_dict` is True, an [`~models.unets.unet_2d_condition.UNet2DConditionOutput`] is returned,
                    otherwise a `tuple` is returned where the first element is the sample tensor.
            """
            # By default samples have to be AT least a multiple of the overall upsampling factor.
            # The overall upsampling factor is equal to 2 ** (# num of upsampling layers).
            # However, the upsampling interpolation output size can be forced to fit any upsampling size
            # on the fly if necessary.
            default_overall_up_factor = 2**self.num_upsamplers

            # upsample size should be forwarded when sample is not a multiple of `default_overall_up_factor`
            forward_upsample_size = False
            upsample_size = None

            for dim in sample.shape[-2:]:
                if dim % default_overall_up_factor != 0:
                    # Forward upsample size to force interpolation output size.
                    forward_upsample_size = True
                    break

            # ensure attention_mask is a bias, and give it a singleton query_tokens dimension
            # expects mask of shape:
            #   [batch, key_tokens]
            # adds singleton query_tokens dimension:
            #   [batch,                    1, key_tokens]
            # this helps to broadcast it as a bias over attention scores, which will be in one of the following shapes:
            #   [batch,  heads, query_tokens, key_tokens] (e.g. torch sdp attn)
            #   [batch * heads, query_tokens, key_tokens] (e.g. xformers or classic attn)
            if attention_mask is not None:
                # assume that mask is expressed as:
                #   (1 = keep,      0 = discard)
                # convert mask into a bias that can be added to attention scores:
                #       (keep = +0,     discard = -10000.0)
                attention_mask = (1 - attention_mask.to(sample.dtype)) * -10000.0
                attention_mask = attention_mask.unsqueeze(1)

            # convert encoder_attention_mask to a bias the same way we do for attention_mask
            if encoder_attention_mask is not None:
                encoder_attention_mask = (
                    1 - encoder_attention_mask.to(sample.dtype)
                ) * -10000.0
                encoder_attention_mask = encoder_attention_mask.unsqueeze(1)

            # 0. center input if necessary
            if self.config.center_input_sample:
                sample = 2 * sample - 1.0

            # 1. time
            t_emb = self.get_time_embed(sample=sample, timestep=timestep)
            emb = self.time_embedding(t_emb, timestep_cond)
            aug_emb = None

            class_emb = self.get_class_embed(sample=sample, class_labels=class_labels)
            if class_emb is not None:
                if self.config.class_embeddings_concat:
                    emb = torch.cat([emb, class_emb], dim=-1)
                else:
                    emb = emb + class_emb

            aug_emb = self.get_aug_embed(
                emb=emb,
                encoder_hidden_states=encoder_hidden_states,
                added_cond_kwargs=added_cond_kwargs,
            )
            if self.config.addition_embed_type == "image_hint":
                aug_emb, hint = aug_emb
                sample = torch.cat([sample, hint], dim=1)

            emb = emb + aug_emb if aug_emb is not None else emb

            if self.time_embed_act is not None:
                emb = self.time_embed_act(emb)

            encoder_hidden_states = self.process_encoder_hidden_states(
                encoder_hidden_states=encoder_hidden_states,
                added_cond_kwargs=added_cond_kwargs,
            )

            # 2. pre-process
            sample = self.conv_in(sample)

            # 2.5 GLIGEN position net
            if (
                cross_attention_kwargs is not None
                and cross_attention_kwargs.get("gligen", None) is not None
            ):
                cross_attention_kwargs = cross_attention_kwargs.copy()
                gligen_args = cross_attention_kwargs.pop("gligen")
                cross_attention_kwargs["gligen"] = {
                    "objs": self.position_net(**gligen_args)
                }

            # 3. down
            # we're popping the `scale` instead of getting it because otherwise `scale` will be propagated
            # to the internal blocks and will raise deprecation warnings. this will be confusing for our users.
            if cross_attention_kwargs is not None:
                cross_attention_kwargs = cross_attention_kwargs.copy()
                lora_scale = cross_attention_kwargs.pop("scale", 1.0)
            else:
                lora_scale = 1.0

            if USE_PEFT_BACKEND:
                # weight the lora layers by setting `lora_scale` for each PEFT layer
                scale_lora_layers(self, lora_scale)

            is_controlnet = (
                mid_block_additional_residual is not None
                and down_block_additional_residuals is not None
            )
            # using new arg down_intrablock_additional_residuals for T2I-Adapters, to distinguish from controlnets
            is_adapter = down_intrablock_additional_residuals is not None
            # maintain backward compatibility for legacy usage, where
            #       T2I-Adapter and ControlNet both use down_block_additional_residuals arg
            #       but can only use one or the other
            if (
                not is_adapter
                and mid_block_additional_residual is None
                and down_block_additional_residuals is not None
            ):
                deprecate(
                    "T2I should not use down_block_additional_residuals",
                    "1.3.0",
                    "Passing intrablock residual connections with `down_block_additional_residuals` is deprecated \
                           and will be removed in diffusers 1.3.0.  `down_block_additional_residuals` should only be used \
                           for ControlNet. Please make sure use `down_intrablock_additional_residuals` instead. ",
                    standard_warn=False,
                )
                down_intrablock_additional_residuals = down_block_additional_residuals
                is_adapter = True

            down_block_res_samples = (sample,)
            for downsample_block in self.down_blocks:

                register_inference_step(downsample_block, inference_step)

                if (
                    hasattr(downsample_block, "has_cross_attention")
                    and downsample_block.has_cross_attention
                ):
                    # For t2i-adapter CrossAttnDownBlock2D
                    additional_residuals = {}
                    if is_adapter and len(down_intrablock_additional_residuals) > 0:
                        additional_residuals["additional_residuals"] = (
                            down_intrablock_additional_residuals.pop(0)
                        )

                    sample, res_samples = downsample_block(
                        hidden_states=sample,
                        temb=emb,
                        encoder_hidden_states=encoder_hidden_states,
                        attention_mask=attention_mask,
                        cross_attention_kwargs=cross_attention_kwargs,
                        encoder_attention_mask=encoder_attention_mask,
                        **additional_residuals,
                    )
                else:
                    sample, res_samples = downsample_block(
                        hidden_states=sample, temb=emb
                    )
                    if is_adapter and len(down_intrablock_additional_residuals) > 0:
                        sample += down_intrablock_additional_residuals.pop(0)

                down_block_res_samples += res_samples

            if is_controlnet:
                new_down_block_res_samples = ()

                for down_block_res_sample, down_block_additional_residual in zip(
                    down_block_res_samples, down_block_additional_residuals
                ):
                    down_block_res_sample = (
                        down_block_res_sample + down_block_additional_residual
                    )
                    new_down_block_res_samples = new_down_block_res_samples + (
                        down_block_res_sample,
                    )

                down_block_res_samples = new_down_block_res_samples

            # 4. mid
            if self.mid_block is not None:

                register_inference_step(self.mid_block, inference_step)

                if (
                    hasattr(self.mid_block, "has_cross_attention")
                    and self.mid_block.has_cross_attention
                ):
                    sample = self.mid_block(
                        sample,
                        emb,
                        encoder_hidden_states=encoder_hidden_states,
                        attention_mask=attention_mask,
                        cross_attention_kwargs=cross_attention_kwargs,
                        encoder_attention_mask=encoder_attention_mask,
                    )
                else:
                    sample = self.mid_block(sample, emb)

                # To support T2I-Adapter-XL
                if (
                    is_adapter
                    and len(down_intrablock_additional_residuals) > 0
                    and sample.shape == down_intrablock_additional_residuals[0].shape
                ):
                    sample += down_intrablock_additional_residuals.pop(0)

            if is_controlnet:
                sample = sample + mid_block_additional_residual

            # 5. up
            for i, upsample_block in enumerate(self.up_blocks):

                register_inference_step(upsample_block, inference_step)

                is_final_block = i == len(self.up_blocks) - 1

                res_samples = down_block_res_samples[-len(upsample_block.resnets) :]
                down_block_res_samples = down_block_res_samples[
                    : -len(upsample_block.resnets)
                ]

                # if we have not reached the final block and need to forward the
                # upsample size, we do it here
                if not is_final_block and forward_upsample_size:
                    upsample_size = down_block_res_samples[-1].shape[2:]

                if (
                    hasattr(upsample_block, "has_cross_attention")
                    and upsample_block.has_cross_attention
                ):
                    sample = upsample_block(
                        hidden_states=sample,
                        temb=emb,
                        res_hidden_states_tuple=res_samples,
                        encoder_hidden_states=encoder_hidden_states,
                        cross_attention_kwargs=cross_attention_kwargs,
                        upsample_size=upsample_size,
                        attention_mask=attention_mask,
                        encoder_attention_mask=encoder_attention_mask,
                    )
                else:
                    sample = upsample_block(
                        hidden_states=sample,
                        temb=emb,
                        res_hidden_states_tuple=res_samples,
                        upsample_size=upsample_size,
                    )

            # 6. post-process
            if self.conv_norm_out:
                sample = self.conv_norm_out(sample)
                sample = self.conv_act(sample)
            sample = self.conv_out(sample)

            if USE_PEFT_BACKEND:
                # remove `lora_scale` from each PEFT layer
                unscale_lora_layers(self, lora_scale)

            if not return_dict:
                return (sample,)

            return UNet2DConditionOutput(sample=sample)

        return forward

    unet.forward = reg_new_forward(unet)


def modify_pipeline(pipe):
    def reg_new_call(self):
        @torch.no_grad
        def call(
            prompt: Union[str, List[str]] = None,
            height: Optional[int] = None,
            width: Optional[int] = None,
            num_inference_steps: int = 50,
            timesteps: List[int] = None,
            sigmas: List[float] = None,
            guidance_scale: float = 7.5,
            negative_prompt: Optional[Union[str, List[str]]] = None,
            num_images_per_prompt: Optional[int] = 1,
            eta: float = 0.0,
            generator: Optional[Union[torch.Generator, List[torch.Generator]]] = None,
            latents: Optional[torch.Tensor] = None,
            prompt_embeds: Optional[torch.Tensor] = None,
            negative_prompt_embeds: Optional[torch.Tensor] = None,
            ip_adapter_image: Optional[PipelineImageInput] = None,
            ip_adapter_image_embeds: Optional[List[torch.Tensor]] = None,
            output_type: Optional[str] = "pil",
            return_dict: bool = True,
            cross_attention_kwargs: Optional[Dict[str, Any]] = None,
            guidance_rescale: float = 0.0,
            clip_skip: Optional[int] = None,
            callback_on_step_end: Optional[
                Union[
                    Callable[[int, int, Dict], None],
                    PipelineCallback,
                    MultiPipelineCallbacks,
                ]
            ] = None,
            callback_on_step_end_tensor_inputs: List[str] = ["latents"],
            **kwargs,
        ):
            r"""
            The call function to the pipeline for generation.

            Args:
                prompt (`str` or `List[str]`, *optional*):
                    The prompt or prompts to guide image generation. If not defined, you need to pass `prompt_embeds`.
                height (`int`, *optional*, defaults to `self.unet.config.sample_size * self.vae_scale_factor`):
                    The height in pixels of the generated image.
                width (`int`, *optional*, defaults to `self.unet.config.sample_size * self.vae_scale_factor`):
                    The width in pixels of the generated image.
                num_inference_steps (`int`, *optional*, defaults to 50):
                    The number of denoising steps. More denoising steps usually lead to a higher quality image at the
                    expense of slower inference.
                timesteps (`List[int]`, *optional*):
                    Custom timesteps to use for the denoising process with schedulers which support a `timesteps` argument
                    in their `set_timesteps` method. If not defined, the default behavior when `num_inference_steps` is
                    passed will be used. Must be in descending order.
                sigmas (`List[float]`, *optional*):
                    Custom sigmas to use for the denoising process with schedulers which support a `sigmas` argument in
                    their `set_timesteps` method. If not defined, the default behavior when `num_inference_steps` is passed
                    will be used.
                guidance_scale (`float`, *optional*, defaults to 7.5):
                    A higher guidance scale value encourages the model to generate images closely linked to the text
                    `prompt` at the expense of lower image quality. Guidance scale is enabled when `guidance_scale > 1`.
                negative_prompt (`str` or `List[str]`, *optional*):
                    The prompt or prompts to guide what to not include in image generation. If not defined, you need to
                    pass `negative_prompt_embeds` instead. Ignored when not using guidance (`guidance_scale < 1`).
                num_images_per_prompt (`int`, *optional*, defaults to 1):
                    The number of images to generate per prompt.
                eta (`float`, *optional*, defaults to 0.0):
                    Corresponds to parameter eta (η) from the [DDIM](https://arxiv.org/abs/2010.02502) paper. Only applies
                    to the [`~schedulers.DDIMScheduler`], and is ignored in other schedulers.
                generator (`torch.Generator` or `List[torch.Generator]`, *optional*):
                    A [`torch.Generator`](https://pytorch.org/docs/stable/generated/torch.Generator.html) to make
                    generation deterministic.
                latents (`torch.Tensor`, *optional*):
                    Pre-generated noisy latents sampled from a Gaussian distribution, to be used as inputs for image
                    generation. Can be used to tweak the same generation with different prompts. If not provided, a latents
                    tensor is generated by sampling using the supplied random `generator`.
                prompt_embeds (`torch.Tensor`, *optional*):
                    Pre-generated text embeddings. Can be used to easily tweak text inputs (prompt weighting). If not
                    provided, text embeddings are generated from the `prompt` input argument.
                negative_prompt_embeds (`torch.Tensor`, *optional*):
                    Pre-generated negative text embeddings. Can be used to easily tweak text inputs (prompt weighting). If
                    not provided, `negative_prompt_embeds` are generated from the `negative_prompt` input argument.
                ip_adapter_image: (`PipelineImageInput`, *optional*): Optional image input to work with IP Adapters.
                ip_adapter_image_embeds (`List[torch.Tensor]`, *optional*):
                    Pre-generated image embeddings for IP-Adapter. It should be a list of length same as number of
                    IP-adapters. Each element should be a tensor of shape `(batch_size, num_images, emb_dim)`. It should
                    contain the negative image embedding if `do_classifier_free_guidance` is set to `True`. If not
                    provided, embeddings are computed from the `ip_adapter_image` input argument.
                output_type (`str`, *optional*, defaults to `"pil"`):
                    The output format of the generated image. Choose between `PIL.Image` or `np.array`.
                return_dict (`bool`, *optional*, defaults to `True`):
                    Whether or not to return a [`~pipelines.stable_diffusion.StableDiffusionPipelineOutput`] instead of a
                    plain tuple.
                cross_attention_kwargs (`dict`, *optional*):
                    A kwargs dictionary that if specified is passed along to the [`AttentionProcessor`] as defined in
                    [`self.processor`](https://github.com/huggingface/diffusers/blob/main/src/diffusers/models/attention_processor.py).
                guidance_rescale (`float`, *optional*, defaults to 0.0):
                    Guidance rescale factor from [Common Diffusion Noise Schedules and Sample Steps are
                    Flawed](https://arxiv.org/pdf/2305.08891.pdf). Guidance rescale factor should fix overexposure when
                    using zero terminal SNR.
                clip_skip (`int`, *optional*):
                    Number of layers to be skipped from CLIP while computing the prompt embeddings. A value of 1 means that
                    the output of the pre-final layer will be used for computing the prompt embeddings.
                callback_on_step_end (`Callable`, `PipelineCallback`, `MultiPipelineCallbacks`, *optional*):
                    A function or a subclass of `PipelineCallback` or `MultiPipelineCallbacks` that is called at the end of
                    each denoising step during the inference. with the following arguments: `callback_on_step_end(self:
                    DiffusionPipeline, step: int, timestep: int, callback_kwargs: Dict)`. `callback_kwargs` will include a
                    list of all tensors as specified by `callback_on_step_end_tensor_inputs`.
                callback_on_step_end_tensor_inputs (`List`, *optional*):
                    The list of tensor inputs for the `callback_on_step_end` function. The tensors specified in the list
                    will be passed as `callback_kwargs` argument. You will only be able to include variables listed in the
                    `._callback_tensor_inputs` attribute of your pipeline class.

            Examples:

            Returns:
                [`~pipelines.stable_diffusion.StableDiffusionPipelineOutput`] or `tuple`:
                    If `return_dict` is `True`, [`~pipelines.stable_diffusion.StableDiffusionPipelineOutput`] is returned,
                    otherwise a `tuple` is returned where the first element is a list with the generated images and the
                    second element is a list of `bool`s indicating whether the corresponding generated image contains
                    "not-safe-for-work" (nsfw) content.
            """

            callback = kwargs.pop("callback", None)
            callback_steps = kwargs.pop("callback_steps", None)

            if callback is not None:
                deprecate(
                    "callback",
                    "1.0.0",
                    "Passing `callback` as an input argument to `__call__` is deprecated, consider using `callback_on_step_end`",
                )
            if callback_steps is not None:
                deprecate(
                    "callback_steps",
                    "1.0.0",
                    "Passing `callback_steps` as an input argument to `__call__` is deprecated, consider using `callback_on_step_end`",
                )

            if isinstance(
                callback_on_step_end, (PipelineCallback, MultiPipelineCallbacks)
            ):
                callback_on_step_end_tensor_inputs = callback_on_step_end.tensor_inputs

            # 0. Default height and width to unet
            height = height or self.unet.config.sample_size * self.vae_scale_factor
            width = width or self.unet.config.sample_size * self.vae_scale_factor
            # to deal with lora scaling and other possible forward hooks

            # 1. Check inputs. Raise error if not correct
            self.check_inputs(
                prompt,
                height,
                width,
                callback_steps,
                negative_prompt,
                prompt_embeds,
                negative_prompt_embeds,
                ip_adapter_image,
                ip_adapter_image_embeds,
                callback_on_step_end_tensor_inputs,
            )

            self._guidance_scale = guidance_scale
            self._guidance_rescale = guidance_rescale
            self._clip_skip = clip_skip
            self._cross_attention_kwargs = cross_attention_kwargs
            self._interrupt = False

            # 2. Define call parameters
            if prompt is not None and isinstance(prompt, str):
                batch_size = 1
            elif prompt is not None and isinstance(prompt, list):
                batch_size = len(prompt)
            else:
                batch_size = prompt_embeds.shape[0]

            device = self._execution_device

            # 3. Encode input prompt
            lora_scale = (
                self.cross_attention_kwargs.get("scale", None)
                if self.cross_attention_kwargs is not None
                else None
            )

            prompt_embeds, negative_prompt_embeds = self.encode_prompt(
                prompt,
                device,
                num_images_per_prompt,
                self.do_classifier_free_guidance,
                negative_prompt,
                prompt_embeds=prompt_embeds,
                negative_prompt_embeds=negative_prompt_embeds,
                lora_scale=lora_scale,
                clip_skip=self.clip_skip,
            )

            # For classifier free guidance, we need to do two forward passes.
            # Here we concatenate the unconditional and text embeddings into a single batch
            # to avoid doing two forward passes
            if self.do_classifier_free_guidance:
                prompt_embeds = torch.cat([negative_prompt_embeds, prompt_embeds])

            if ip_adapter_image is not None or ip_adapter_image_embeds is not None:
                image_embeds = self.prepare_ip_adapter_image_embeds(
                    ip_adapter_image,
                    ip_adapter_image_embeds,
                    device,
                    batch_size * num_images_per_prompt,
                    self.do_classifier_free_guidance,
                )

            # 4. Prepare timesteps
            timesteps, num_inference_steps = retrieve_timesteps(
                self.scheduler, num_inference_steps, device, timesteps, sigmas
            )

            # 5. Prepare latent variables
            num_channels_latents = self.unet.config.in_channels
            latents = self.prepare_latents(
                batch_size * num_images_per_prompt,
                num_channels_latents,
                height,
                width,
                prompt_embeds.dtype,
                device,
                generator,
                latents,
            )

            # 6. Prepare extra step kwargs. TODO: Logic should ideally just be moved out of the pipeline
            extra_step_kwargs = self.prepare_extra_step_kwargs(generator, eta)

            # 6.1 Add image embeds for IP-Adapter
            added_cond_kwargs = (
                {"image_embeds": image_embeds}
                if (ip_adapter_image is not None or ip_adapter_image_embeds is not None)
                else None
            )

            # 6.2 Optionally get Guidance Scale Embedding
            timestep_cond = None
            if self.unet.config.time_cond_proj_dim is not None:
                guidance_scale_tensor = torch.tensor(self.guidance_scale - 1).repeat(
                    batch_size * num_images_per_prompt
                )
                timestep_cond = self.get_guidance_scale_embedding(
                    guidance_scale_tensor,
                    embedding_dim=self.unet.config.time_cond_proj_dim,
                ).to(device=device, dtype=latents.dtype)

            # 7. Denoising loop
            num_warmup_steps = (
                len(timesteps) - num_inference_steps * self.scheduler.order
            )
            self._num_timesteps = len(timesteps)
            with self.progress_bar(total=num_inference_steps) as progress_bar:
                for i, t in enumerate(timesteps):
                    if self.interrupt:
                        continue

                    # expand the latents if we are doing classifier free guidance
                    latent_model_input = (
                        torch.cat([latents] * 2)
                        if self.do_classifier_free_guidance
                        else latents
                    )
                    latent_model_input = self.scheduler.scale_model_input(
                        latent_model_input, t
                    )

                    # predict the noise residual
                    noise_pred = self.unet(
                        latent_model_input,
                        t,
                        i + 1,  # Pass inference step start from zero
                        encoder_hidden_states=prompt_embeds,
                        timestep_cond=timestep_cond,
                        cross_attention_kwargs=self.cross_attention_kwargs,
                        added_cond_kwargs=added_cond_kwargs,
                        return_dict=False,
                    )[0]

                    # perform guidance
                    if self.do_classifier_free_guidance:
                        noise_pred_uncond, noise_pred_text = noise_pred.chunk(2)
                        noise_pred = noise_pred_uncond + self.guidance_scale * (
                            noise_pred_text - noise_pred_uncond
                        )

                    if self.do_classifier_free_guidance and self.guidance_rescale > 0.0:
                        # Based on 3.4. in https://arxiv.org/pdf/2305.08891.pdf
                        noise_pred = rescale_noise_cfg(
                            noise_pred,
                            noise_pred_text,
                            guidance_rescale=self.guidance_rescale,
                        )

                    # compute the previous noisy sample x_t -> x_t-1
                    latents = self.scheduler.step(
                        noise_pred, t, latents, **extra_step_kwargs, return_dict=False
                    )[0]

                    if callback_on_step_end is not None:
                        callback_kwargs = {}
                        for k in callback_on_step_end_tensor_inputs:
                            callback_kwargs[k] = locals()[k]
                        callback_outputs = callback_on_step_end(
                            self, i, t, callback_kwargs
                        )

                        latents = callback_outputs.pop("latents", latents)
                        prompt_embeds = callback_outputs.pop(
                            "prompt_embeds", prompt_embeds
                        )
                        negative_prompt_embeds = callback_outputs.pop(
                            "negative_prompt_embeds", negative_prompt_embeds
                        )

                    # call the callback, if provided
                    if i == len(timesteps) - 1 or (
                        (i + 1) > num_warmup_steps
                        and (i + 1) % self.scheduler.order == 0
                    ):
                        progress_bar.update()
                        if callback is not None and i % callback_steps == 0:
                            step_idx = i // getattr(self.scheduler, "order", 1)
                            callback(step_idx, t, latents)

            if not output_type == "latent":
                image = self.vae.decode(
                    latents / self.vae.config.scaling_factor,
                    return_dict=False,
                    generator=generator,
                )[0]
                image, has_nsfw_concept = self.run_safety_checker(
                    image, device, prompt_embeds.dtype
                )
            else:
                image = latents
                has_nsfw_concept = None

            if has_nsfw_concept is None:
                do_denormalize = [True] * image.shape[0]
            else:
                do_denormalize = [not has_nsfw for has_nsfw in has_nsfw_concept]

            image = self.image_processor.postprocess(
                image, output_type=output_type, do_denormalize=do_denormalize
            )

            # Offload all models
            self.maybe_free_model_hooks()

            if not return_dict:
                return (image, has_nsfw_concept)

            return StableDiffusionPipelineOutput(
                images=image, nsfw_content_detected=has_nsfw_concept
            )

        return call

    pipe.call = reg_new_call(pipe)


def apply_CA_ToMe(pipe, cache_conf, r):
    """
    Sets up a diffusion pipeline with Cached Adaptive ToMe optimizations.

    Args:
        pipe: StableDiffusionPipeline, The diffusion pipeline to modify
        cache_conf: List[int], array of timestep checkpoints where new merge indices are computed.
                   These checkpoints specify when the CA-ToMe merging pattern is recalculated.
        r: float, cosine similarity threshold that determines whether tokens should be merged.
           Higher values mean tokens need to be more similar to be merged.

    Returns:
        The modified pipeline with CA-ToMe optimizations applied

    This function applies three modifications to the pipeline:
    1. Modifies the UNet to support cached adaptive ToMe during inference
    2. Modifies the pipeline's forward pass to handle CA-ToMe
    3. Applies the CA-ToMe patch with the specified configuration.
    """
    modify_unet(pipe.unet)
    modify_pipeline(pipe)
    apply_patch(pipe, cache_conf, r)
    return pipe



class CacheConf:
    """
    Cache configurations (checkpoints)
    These are the checkpoints that we used in our experiments,
    documented in Table II of our paper. The corresponding results for each
    configuration can be found in Table III.

    CONFIG_3 demonstrated the best performance in our experiments.
    For detailed analysis, please refer to our paper.

    Example usage:
        from ca_tome.utils import CacheConf
        conf = CacheConf.ADAPTIVE
    """
    CONFIG_1 = [0, 1, 2, 3, 5, 10, 15, 25, 35]
    CONFIG_2 = [0, 10, 11, 12, 15, 20, 25, 30, 35, 45] 
    CONFIG_3 = [0, 8, 11, 13, 20, 25, 30, 35, 45, 46, 47, 48, 49]  # Best performing CA-ToMe configuration
    CONFIG_4 = [0, 9, 13, 14, 15, 28, 29, 32, 36, 45]
    ADAPTIVE = [0, 1, 5, 7, 10, 12, 15, 35, 40, 45, 46, 47, 48, 49, 50, 51]
    CONFIG_Five = [0, 5, 10, 15, 20, 25, 30, 35, 40, 45, 50]



import torch


def isinstance_str(x: object, cls_name: str):
    """
    Checks whether x has any class *named* cls_name in its ancestry.
    Doesn't require access to the class's implementation.

    Useful for CA-ToMe patching!
    """

    for _cls in x.__class__.__mro__:
        if _cls.__name__ == cls_name:
            return True

    return False


def init_generator(device: torch.device, fallback: torch.Generator = None):
    """
    Forks the current default random generator given device.
    Used by CA-ToMe for consistent random token selection.
    """
    if device.type == "cpu":
        return torch.Generator(device="cpu").set_state(torch.get_rng_state())
    elif device.type == "cuda":
        return torch.Generator(device=device).set_state(torch.cuda.get_rng_state())
    else:
        if fallback is None:
            return init_generator(torch.device("cpu"))
        else:
            return fallback
